"""
This module provides implementations based on max-min sampling for various
aspects of cluster generation.

CLASSES AND METHODS
	
	MaxMinClusters : data generator based on max-min sampling
		__init__(self, n_clusters, n_dim, n_samples, ...)

	MaxMinCov : sample cluster shapes (covariance structures)
		__init__(self, ref_aspect, aspect_maxmin, radius_maxmin)
		make_cluster_aspects(self, n_clusters)
		make_cluster_radii(self, n_clusters, ref_radius, n_dim)
		make_axis_sd(self, n_axes, sd, aspect)
		make_cov(self, clusterdata)

	MaxMinBal : sample number of data points per cluster
		__init__(self, imbal_ratio)
		float_to_int(self, float_class_sz, n_samples)
		make_class_sizes(self, clusterdata)

	maxmin_sampler(n_samples, ref, min_val, maxmin_ratio, f_constrain)

"""

from .core import ClusterData, CovGeom, ClassBal
from .centers import BoundedSepCenters
from .distributions import GaussianData, ExpData, tData
import numpy as np
import scipy.stats as stats

class MaxMinClusters(ClusterData):
	"""
	Data generator based on max-min sampling. Generate data sets that share
	geometric properties based on max-min ratios. This is the default 
	implementation of ClusterData. 

	You can specify the sampling mechanism for each of the following geometric
	parameters: cluster radius, cluster aspect ratio and class size.


	Geometric attributes
	--------------------
	Cluster radius : float, >0
		Geometric mean of the standard deviations along a cluster's principal
		axes (eigenvectors of covariance matrix)
	Cluster aspect ratio : float, >=1
		Ratio between maximum and minimum standard deviations across a cluster's
		principal axes (eigenvectors of covariance matrix).
	Class size : int
		Number of data points in a cluster


	Configuring geometric attributes
	--------------------------------
	radius_maxmin : float, >=1
		Ratio between maximum and minimum radii across all clusters
	aspect_maxmin : float, >=1
		Ratio between maximum and minimum aspect ratios across all clusters
	aspect_ref : float, >=1
		Reference aspect ratio (aspect ratio a "typical" cluster should have)
	imbal_maxmin : float, >=1
		Ratio between highest and lowest class sizes across clusters


	Attributes
	----------
	All attributes are inherited from the class ClusterData. However, a
	MaxMinClusters object uses specific implementations of the following 
	attributes:

	center_geom : centers.BoundedSepCenters
		Sample cluster centers that satisfy separation constraints
	cov_geom : MaxMinCov
		Sample cluster shapes (covariance structures) with max-min sampling
	class_bal : MaxMinBal
		Sample class sizes (data points in each cluster) with max-min sampling
	data_dist : GaussianData, ExpData, or tData
		Data distribution is chosen at initialization (the dist argument)


	Methods
	-------
	__init__(self, n_clusters, n_dim, n_samples, ...)

	"""

	def __init__(self, n_clusters=6, n_dim=2, n_samples=500, imbal_maxmin=2,
				 aspect_maxmin=2, radius_maxmin=3, alpha_max=0.05, alpha_min=1e-3, 
				 aspect_ref=1.5, packing=0.1,dist='gaussian',df=1, scale=1.0):
		"""
		Create a MaxMinClusters object. 

		See the documentation of class MaxMinClusters for more information about
		the following geometric parameters: cluster radius, cluster aspect ratio,
		and class size.

		Parameters
		----------
		self : MaxMinClusters
			This instance of MaxMinClusters
		n_clusters : int
			Number of clusters to generate.
		n_dim : int
			Dimensionality of the data set to generate.
		n_samples : int
			Total number of samples to generate.
		radius_maxmin : float, >=1
			Ratio between the maximum and minimum cluster radii
		aspect_maxmin : float, >=1
			Ratio between maximum and minimum cluster aspect ratios
		aspect_ref : float, >=1
			Reference (typical) aspect ratio for clusters
		imbal_maxmin : float, >=1
			Ratio between highest and lowest class sizes across clusters
		min_sep : float, >0
			Minimum separation between clusters. For Gaussian clusters, a 
			value of 1.0 indicates that two clusters are just touching (see 
			centers.BoundedSepCenters for more information)
		max_sep : float, >0
			Maximum separation between clusters. For Gaussian clusters, a 
			value of 1.0 indicates that two clusters are just touching (see 
			centers.BoundedSepCenters for more information)
		packing : float between 0 and 1
			Ratio between total volume of clusters to sampling volume (see
			centers.BoundedSepCenters for more information)
		dist : str
			String indicating which probability distribution to use for drawing
			data. There are currently three possibilities: 'gaussian' for 
			Gaussian data, 't' for standard t-distributed data, and 'exp' for 
			double exponentially distributed data.
		df : int
			Degrees of freedom in Student's t distribution to use when
			sampling t-distributed clusters (applicable when dist='t')
		scale : float
			Reference length scale for generated data

		Returns
		-------
		out : MaxMinClusters
			Data generator for sampling synthetic datasets with desired 
			geometric characteristics

		"""

		cov_geom = MaxMinCov(ref_aspect=aspect_ref, aspect_maxmin=aspect_maxmin, 
							 radius_maxmin=radius_maxmin)
		center_geom = BoundedSepCenters(alpha_max=alpha_max,alpha_min=alpha_min, 
										packing=packing)
		class_bal = MaxMinBal(imbal_ratio=imbal_maxmin)

		if dist=='t':
			data_dist = tData(df=df)
		elif dist=='exp':
			data_dist = ExpData()
		elif dist=='gaussian':
			data_dist = GaussianData()
		else:
			raise ValueError("Distribution not found. Use dist='gaussian' " +
							 "for Gaussian data, dist='t' for t-distributed data," + 
							 " or dist='exp' for exponentially distributed data.")

		# in line below, used to be super().__init__
		ClusterData.__init__(self, n_clusters,n_dim,n_samples,class_bal,cov_geom,
							 center_geom,data_dist,scale)


class MaxMinCov(CovGeom):
	"""
	Defines cluster shapes by setting ratios between maximum and minimum values
	of geometric parameters. The algorithms use pairwise max-min sampling.

	See documentation of class MaxMinClusters for more information.

	Attributes
	----------
	ref_aspect : float, >= 1            
		Reference aspect ratio for each cluster.
	aspect_maxmin : float, >= 1
		Desired ratio between maximum and minimum aspect ratios among clusters.
	radius_max_min : float, >= 1
		Desired ratio between maximum and minimum cluster radius.

	Methods
	-------
	__init__(self, ref_aspect, aspect_maxmin, radius_maxmin)
	make_cluster_aspects(self, n_clusters)
	make_cluster_radii(self, n_clusters, ref_radius, n_dim)
	make_axis_sd(self, n_axes, sd, aspect)
	make_cov(self, clusterdata, ...)

	"""
	
	def __init__(self, ref_aspect, aspect_maxmin, radius_maxmin):
		"""
		Constructs a MaxMinCov object.

		Parameters
		----------
		self : MaxMinCov
			This instance of MaxMinCov
		ref_aspect : float, >= 1            
			Reference aspect ratio for each cluster.
		aspect_maxmin : float, >= 1
			Desired ratio between maximum and minimum aspect ratios among 
			clusters.
		radius_max_min : float, >= 1
			Desired ratio between maximum and minimum cluster radius.

		"""
		self.ref_aspect = ref_aspect
		self.aspect_maxmin = aspect_maxmin
		self.radius_maxmin = radius_maxmin
	

	def make_cluster_aspects(self, n_clusters,seed=None):
		"""
		Generates aspect ratios (ratio between standard deviations along longest
		and shortest axes) for all clusters.

		Parameters
		----------
		self : MaxMinCov
			This instance of MaxMinCov
		n_clusters : int
			The number of clusters.

		Returns
		-------
		out : ndarray
			The aspect ratios for each cluster.
		"""

		min_aspect = 1 + (self.ref_aspect-1)/np.sqrt(self.aspect_maxmin)
		f = lambda a: ((self.ref_aspect-1)**2)/a
		return 1+maxmin_sampler(n_clusters, self.ref_aspect-1, min_aspect-1, self.aspect_maxmin, f,
								seed=seed)

		
	def make_cluster_radii(self, n_clusters, ref_radius, n_dim, seed=None):
		""" 
		Sample cluster radii with pairwise max-min sampling.

		The radius of a cluster is the geometric mean of the standard deviations along 
		the principal axes. Cluster radii are sampled such that the arithmetic mean of
		cluster volumes (cluster radius to the n_dim power) equals the reference volume
		(ref_radius to the n_dim power). The minimum and maximum radii are chosen so that
		the arithmetic average of the corresponding volumes equals the reference volume.

		Parameters
		----------
		self : MaxMinCov
			This instance of MaxMinCov
		n_clusters : int
			Number of clusters
		ref_radius : float
			Reference radius for all clusters
		n_dim : int
			Dimensionality of the data

		Returns
		-------
		out : ndarray
			Cluster radii
		"""
		log_min_radius = np.log(ref_radius) - (1/2)*np.log(self.radius_maxmin)
		f = lambda log_r: 2*np.log(ref_radius) - log_r

		#min_radius = (2*(ref_radius**n_dim)/(1 + self.radius_maxmin**n_dim))**(1/n_dim)
		#f = lambda r: (2*(ref_radius**n_dim) - (r**n_dim))**(1/n_dim)

		log_max_radius = np.log(ref_radius) + (1/2)*np.log(self.radius_maxmin)
		maxmin_log_ratio = log_max_radius/log_min_radius

		return np.exp(maxmin_sampler(n_clusters, np.log(ref_radius), log_min_radius, maxmin_log_ratio, f,
							  seed=seed))
	

	def make_axis_sd(self, n_axes, sd, aspect, seed=None):
		"""
		Sample standard deviations for the principal axes of a single cluster.

		Parameters
		----------
		self : MaxMinCov
			This instance of MaxMinCov
		n_axes : int
			Number of principal axes of this cluster, same as dimensionality
		sd : float
			Overall standard deviation of this cluster (geometric mean of 
			standard deviations)
		aspect : float
			Desired ratio between maximum and minimum standard deviations
			across all principal axes

		Returns
		-------
		out : ndarray
			Standard deviations along principal axes of this cluster

		"""
		min_sd = sd/np.sqrt(aspect)
		f = lambda s: (sd**2)/s
		return maxmin_sampler(n_axes, sd, min_sd, aspect, f, seed=seed)
		

	def make_cov(self, clusterdata, seed=None):
		"""
		Compute covariance structure (cluster shape) for each cluster.

		Parameters
		----------
		self : MaxMinCov
			This instance of MaxMinCov
		clusterdata : ClusterData
			Specifies the number of clusters and other parameters

		Returns
		-------
		out : tuple (axis, sd, cov, cov_inv), where

			axis : list of ndarray
				The i-th element stores principal axes of the i-th cluster
			sd : list of ndarray
				The i-th element stores standard deviations along the principal
				axes of the i-th cluster
			cov : list of ndarray
				The i-th element stores covariance matrix of the i-th cluster
			cov_inv : list of ndarray
				The i-th element stores inverse covariance matrix of the i-th
				cluster

			Matching indices of the output lists refer to the same cluster.

		"""
		if seed:
			np.random.seed(seed)

		axis = list()
		sd = list()
		cov = list()
		cov_inv = list()

		n_clusters = clusterdata.n_clusters
		n_dim = clusterdata.n_dim
		scale = clusterdata.scale
		
		cluster_radii = self.make_cluster_radii(n_clusters, scale, n_dim, seed=seed+2*n_clusters if seed else seed)
		cluster_aspects = self.make_cluster_aspects(n_clusters,seed=seed+3*n_clusters if seed else seed)
		
		for clust in range(n_clusters):
			# compute principal axes for cluster
			axes = self.make_orthonormal_axes(n_dim, n_dim, seed=seed+clust if seed else seed)
			axis_sd = self.make_axis_sd(n_dim, cluster_radii[clust], cluster_aspects[clust],
										seed=seed+clust if seed else seed)

			axis.append(axes)
			sd.append(axis_sd)

			# consider not constructing cov, cov_inv here; instead do it only on request
			cov.append(np.transpose(axes) @ np.diag(axis_sd**2) @ axes)
			cov_inv.append(np.transpose(axes) @ np.diag(1/axis_sd**2) @ axes)

		out = (axis, sd, cov, cov_inv)
			
		return out


class MaxMinBal(ClassBal):
	"""
	Generate class sizes (number of data points in each cluster) with
	pairwise max-min sampling.

	Pairwise max-min sampling uses the average class size as the reference
	value. The sampled class sizes sum to the desired total number of samples
	as specified by the relevant ClusterData object (an argument in the method
	MaxMinBal.make_class_sizes).

	Attributes
	----------
	imbal_ratio : float, >=1
		Desired ratio between largest and smallest class size

	Methods
	-------
	__init__(self, imbal_ratio)
	make_class_sizes(self, clusterdata)

	"""

	def __init__(self, imbal_ratio):
		"""
		Instantiate a MaxMinBal object.

		See the documentation of class MaxMinBal for more information.

		Parameters
		----------
		self : MaxMinBal
			This instance of MaxMinBal
		imbal_ratio : float, >=1
			Desired ratio between largest and smallest class size 

		Returns
		-------
		out : MaxMinBal
			MaxMinBal object for sampling class sizes
		"""

		self.imbal_ratio = imbal_ratio

	def float_to_int(self, float_class_sz, n_samples):
		"""
		Convert float class sizes to integer class sizes 
		while ensuring 1) that each class size is at least
		1 and 2) that the sum of class sizes is n_samples.

		Parameters
		----------
		self : MaxMinBal
			This instance of MaxMinBal
		float_class_sz : ndarray, dtype=float
			Approximate class sizes, not necessarily integer
		n_samples : int
			Desired total number of samples (sum of class sizes)

		Returns
		-------
		out : ndarray, dtype=int
			Class sizes (number of data points in each cluster)
		"""
		# round float class sizes and add 1, then sort
		class_sz = 1 + np.sort(np.round(float_class_sz))
		# start by shrinking the highest class sizes
		class2shrink_idx = len(class_sz) - 1
		while (np.sum(class_sz) > n_samples):
			if (class_sz[class2shrink_idx] > 1):
				class_sz[class2shrink_idx] -= 1
				class2shrink_idx -= 1
			else:
				class2shrink_idx = len(class_sz) - 1
		return class_sz.astype(int)

	def make_class_sizes(self, clusterdata, seed=None):
		"""
		Sample class size (number of data points) for each cluster with
		pairwise max-min sampling. 

		See the documentation of class MaxMinBal for more information.

		Parameters
		----------
		self : MaxMinBal
			This instance of MaxMinBal
		clusterdata : ClusterData
			The underlying data generator

		Returns
		-------
		out : ndarray
			Class sizes (number of data points in each cluster)

		"""
		n_samples = clusterdata.n_samples
		n_clusters = clusterdata.n_clusters

		# Set average class size as the reference size.
		ref_class_sz = n_samples/n_clusters

		# Determine minimum class size by requiring average of minimum and maximum
		# class sizes to be the reference size.
		min_class_sz = 2*ref_class_sz/(1 + self.imbal_ratio)

		# Set pairwise sampling constraint to ensure sample sizes add to n_samples.
		f = lambda s: (2*ref_class_sz - s)

		# compute float class size estimates
		float_class_sz = maxmin_sampler(n_clusters, ref_class_sz, 
										min_class_sz, self.imbal_ratio, f, 
										seed=seed)

		# transform float class size estimates into integer class sizes
		class_sz = self.float_to_int(float_class_sz, n_samples)

		return class_sz


def maxmin_sampler(n_samples, ref, min_val, maxmin_ratio, f_constrain, seed=None):
	"""
	Generates samples around a reference value, with a fixed ratio between the 
	maximum and minimum sample. Sample pairwise to enforce a further constraint 
	on the samples. For example, the geometric mean of the samples can be 
	specified.

	Parameters
	----------
	n_samples : int
	ref : float
	min_val : float
	maxmin_ratio : float
	f_constrain : function

	Returns
	-------
	out : ndarray

	"""
	np.random.seed(seed)

	if (maxmin_ratio == 1) or (min_val == 0):
		out = np.full(n_samples, fill_value=ref)
		return out

	max_val = min_val * maxmin_ratio
	
	if (n_samples > 2):
		# Besides min_val and max_val, only need n-2 samples
		n_gotta_sample = n_samples-2 
		samples = np.full(n_gotta_sample, fill_value=float(ref))
		# Sample according to triangular distribution with endpoints given by min_val
		# and max_val, and mode given by ref. Sample pairwise. The first sample in each
		# pair is generated randomly, and the second sample is calculated from the first.
		while (n_gotta_sample >= 2):
			samples[n_gotta_sample-1] = np.random.triangular(left=min_val, mode=ref, 
																right=max_val)
			samples[n_gotta_sample-2] = f_constrain(samples[n_gotta_sample-1])
			n_gotta_sample -= 2
		out = np.concatenate([[min_val], np.sort(samples), [max_val]])
	elif (n_samples == 2):
		out = np.array([min_val, max_val])
	elif (n_samples == 1):
		out = np.array([ref])

	permuted_out = out[np.random.permutation(out.size)]
	return permuted_out


"""
This module provides several probability distributions for drawing synthetic
data for clusters.

CLASSES AND METHODS

	GaussianData : draw Gaussian data for cluster
		__init__(self)
		sample_cluster(self,class_size,mean,axes,sd)

	ExpData : draw doubly exponentially distributed data for cluster
		__init__(self)
		sample_cluster(self,class_size,mean,axes,sd)

	tData : draw Student t-distributed data for cluster
		__init__(self)
		sample_cluster(self,class_size,mean,axes,sd)

"""

from .core import DataDist
import numpy as np


class GaussianData(DataDist):
	"""
	Draw Gaussian data for a cluster.

	Attributes
	----------
	None

	Methods
	-------
	__init__(self, ...)
	sample_cluster(self, class_size, mean, axes, sd, ...)

	"""

	def __init__(self):
		"""
		Instantiate a GaussianData object.

		"""
		pass

	def sample_cluster(self,class_size,mean,axes,sd,seed=None):
		"""
		Sample a Gaussian cluster.

		class_size : int
			The number of data points in this cluster.
		mean : ndarray
			Cluster of this cluster.
		axes : list of ndarray
			Principle axes of this cluster.
		sd : list of float
			Standard deviations of this cluster's principal axes.
		"""

		# assemble covariance matrix
		cov = np.transpose(axes) @ np.diag(sd**2) @ axes
		# sample data
		np.random.seed(seed)
		X = np.random.multivariate_normal(mean=mean, 
									cov=cov, size=class_size)
		return X


class ExpData(DataDist):
	"""
	Draw exponentially distributed synthetic data for a cluster.

	Attributes
	----------
	None

	Methods
	-------
	__init__(self, ...)
	sample_cluster(self, class_size, mean, axes, sd, ...)

	"""

	def __init__(self):
		"""
		Instantiate an ExpData object.

		"""
		pass

	def sample_cluster(self,class_size, mean,axes,sd,seed=None):
		"""
		Sample an exponentially distributed cluster.

		Sampling is performed using spherical coordinates. First, a random
		point on the unit sphere is sampled. This point is then scaled by 
		a draw from an exponential distribution with standard deviation 1.
		Finally, the point is stretched or compressed along each principal
		axis of the cluster by the corresponding standard deviation.

		class_size : int
			The number of data points in this cluster.
		mean : ndarray
			Cluster of this cluster.
		axes : list of ndarray
			Principle axes of this cluster.
		sd : list of float
			Standard deviations of this cluster's principal axes.
		"""

		np.random.seed(seed)

		# each row of axes is an axis
		n_axes = axes.shape[0]

		# sample on the unit sphere
		X = np.random.multivariate_normal(mean=np.zeros(n_axes), cov=np.eye(n_axes), 
										  size=class_size)
		X /= np.sqrt(np.sum(X**2,axis=1))[:,np.newaxis]

		# scale sampled unit vectors by draws from standard exponential dist
		X *= np.random.exponential(scale=1,size=class_size)[:,np.newaxis]

		# stretch/compress samples along principal axes
		X = X @ np.diag(sd) @ axes

		# translate sampled data to the desired mean
		X = X + mean[np.newaxis,:]

		return X


class tData(DataDist):
	"""
	Draw t-distributed data for a cluster.

	Attributes
	----------
	df : int
		Degrees of freedom in Student t distribution. Low values (e.g. 1) lead
		to heavy tails. For very high values, the distribution becomes Gaussian.

	Methods
	-------
	__init__(self, ...)
	sample_cluster(self, class_size, mean, axes, sd, ...)

	"""

	def __init__(self, df=1):
		"""
		Instantiate a tData object.

		Parameters
		----------
		self : tData
			This instance of tData
		df : int
			Degrees of freedom in Student t distribution

		"""
		self.df = df

	def sample_cluster(self,class_size, mean,axes,sd,n_empirical_quantile=1e+6, seed=None):
		"""
		Sample a t-distributed cluster.

		Sampling is performed using spherical coordinates. First, a random
		point on the unit sphere is sampled. This point is then scaled by 
		a draw from a Student t distribution with self.df degrees of freedom,
		normalized by the median of the absolute value of the same t 
		distribution. Finally, the sampled point is stretched or compressed 
		along each principal axis of the cluster by the corresponding standard
		deviation.

		class_size : int
			The number of data points in this cluster.
		mean : ndarray
			Cluster of this cluster.
		axes : list of ndarray
			Principle axes of this cluster.
		sd : list of float
			Standard deviations of this cluster's principal axes.
		n_empirical_quantile : int, default=1e+6
			Number of samples used to approximate the median of the absolute
			value of the chosen t distribution
		"""

		np.random.seed(seed)

		# compute median for absolute value of t distribution
		n = n_empirical_quantile
		abs_t_quantile = np.quantile(np.abs(np.random.standard_t(df=self.df,size=int(n))),q=0.68)

		# each row of axes is an axis
		n_axes = axes.shape[0]

		# sample on the unit sphere
		X = np.random.multivariate_normal(mean=np.zeros(n_axes), cov=np.eye(n_axes), 
										  size=class_size)
		X /= np.sqrt(np.sum(X**2,axis=1))[:,np.newaxis]

		# scale sampled unit vectors by draw from normalized t-distribution 
		scaling = 1/abs_t_quantile
		X *= np.random.standard_t(df=self.df, size=class_size)[:,np.newaxis] * scaling

		# stretch/compress along principal axes
		X = X @ np.diag(sd) @ axes

		# mean-shift X
		X = X + mean[np.newaxis,:]

		return X


"""
This module provides the core object-oriented framework for generating 
synthetic data sets with clusters. The classes contained here are mainly 
abstract superclasses. They require subclasses to concretely implement 
much of the specified functionality.

CLASSES AND METHODS

	ClusterData : top-level object for generating data
		__init__(self, n_clusters, n_dim, n_samples, class_bal, cov_geom, 
				 center_geom, data_dist, ...)
		generate_data(self, ...)
		to_csv(self, filename)

	CovGeom : sample cluster shapes
		__init__(self)
		make_cov(self, clusterdata)
		make_orthonormal_axes(self, n_dim, n_axes)

	CenterGeom : place cluster centers
		__init__(self)
		make_centers(self, clusterdata)

	ClassBal : sample number of points for each cluster
		__init__(self)
		make_class_sizes(self, clusterdata)

	DataDist : draw synthetic data for each cluster
		__init__(self)
		sample(self, clusterdata)
		sample_cluster(self, cluster_size, mean, axes, sd)
"""

import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as stats
from scipy.spatial.distance import mahalanobis

class ClusterData:
	"""
	Instances of this class are data generators for sampling synthetic data sets
	with clusters. All generated data sets are different, but possess similar
	geometric characteristics, as specified by an instance of this class.

	Attributes
	----------
	center_geom : CenterGeom
		Algorithm for placing cluster centers
	centers : ndarray
		Cluster centers arranged in a matrix (each row is a center)
	class_bal : ClassBal
		Algorithm for sampling class sizes (number of points in each cluster)
	class_sizes : ndarray, dtype=int
		Number of points in each cluster
	cluster_axes : list of ndarray
		The i-th element stores the principal axes of the i-th cluster,
		arranged in matrix form so that each row is an axis
	cluster_sd : list of ndarray
		The i-th element stores the standard deviations along the principal
		axes of the i-th cluster
	cov : list of ndarray
		The i-th element stores the covariance matrix of the i-th cluster
	cov_geom : CovGeom
		Algorithm for sampling cluster shapes (covariance structures)
	cov_inv : list of ndarray
		The i-th element stores the inverse covariance matrix of the i-th 
		cluster
	data : ndarray or None
		Data set sampled most recently with ClusterData.generate_data()
	data_dist : DataDist
		Distribution for drawing synthetic data
	labels : ndarray, dtype=int
		Cluster labels / class labels
	n_clusters : int
		Number of clusters to sample
	n_dim : int
		Dimensionality of the data
	n_samples : int 
		Total number of samples to generate in each data set
	scale : float
		Reference length scale for generated data

	Methdods
	--------
	__init__(self, n_clusters, n_dim, n_samples, class_bal, cov_geom, 
			 center_geom, data_dist, ...)
	generate_data(self, ...)
	to_csv(self, filename)

	"""
		
	def __init__(self, n_clusters, n_dim, n_samples, class_bal, cov_geom, 
		center_geom, data_dist, scale=1.0):
		"""
		Instantiate a ClusterData object.

		Parameters
		----------
		n_clusters : int
			Number of clusters
		n_dim : int
			Dimensionality of data points
		n_samples : int
			Total number of data points 
		class_bal : ClassBal object
			Samples number of points per cluster
		cov_geom : CovGeom object
			Samples cluster shapes (covariance structures)
		center_geom : CenterGeom object
			Places cluster centers
		data_dist : DataDist object
			Draws synthetic data for each cluster
		scale : float, optional
			Sets reference length scale for generated data

		Returns
		-------
		self : ClusterData
			An instance of ClusterData

		"""
		self.n_clusters = n_clusters
		self.n_dim = n_dim
		self.n_samples = n_samples
		self.class_bal = class_bal
		self.cov_geom = cov_geom
		self.center_geom = center_geom
		self.data_dist = data_dist
		self.data = None
		self.labels = None
		self.centers = None
		self.class_sizes = None
		self.cov = None
		self.cov_inv = None
		self.cluster_axes = None
		self.cluster_sd = None
		self.scale = scale

	def to_csv(self, filename):
		"""
		Write the most recently generated data to a csv file.

		Parameters
		----------
		self : ClusterData
			The underlying data generator
		filename : str
			Filepath for storing data

		Returns
		-------
		None

		Side effects
		------------
		.csv file is created at the location indicated by filename

		"""
		if (self.data is None):
			raise Exception('No data has been generated. Use ClusterData.generate_data to generate data.')
		else:
			np.savetxt(filename, self.data, delimiter=',')


	def modify_cov_structure(self,cov,cov_inv, cluster_axes, cluster_sd):
		"""
		Modify the covariance structure of a ClusterData object.
		
		Need to specify ALL attributes of the covariance structure.
		"""
		self.cov = cov
		self.cov_inv = cov_inv
		self.cluster_axes = cluster_axes
		self.cluster_sd = cluster_sd


	def draw_model(self,s=30,alpha_max=0.125, sd_max_bare=2,sd_max_density=4, mode='bare', h=0.1,dr=0.1,
				   cluster_labels=True):
		"""
		Plot a sketch of the geometry underlying a probabilistic mixture model sampled by a ClusterData object.
		
		Parameters
		----------
		clusterdata : ClusterData object
		s : float
		alpha_max :
		sd_max_bare :
		sd_max_density : 
		mode : str
			Specify the desired type of drawing. The options are
			   'bare' : draw each cluster as a central point and elliptical contour,
						which is located two standard deviations away from the mean
			   'density' : draw each cluster as a probability density that fades out
						   gradually until the elliptical contour four standard
						   deviations away from the mean
			   'arrows' : as in 'bare', except that arrows are added along the principal axes
						NOT IMPLEMENTED YET
		"""

		centers = self.centers
		sd = self.cluster_sd
		axes = self.cluster_axes

		for i in range(centers.shape[0]):
			# plot cluster center
			x_c = centers[i,0]
			y_c = centers[i,1]
			plt.scatter(x_c,y_c,c='k',s=s)
			if cluster_labels:
				plt.text(x_c,y_c,' ' + str(i),fontsize=16,va='center')
			
			if mode=='bare':
				# plot only a single elliptical contour and the cluster center
				
				r = sd_max_bare
				x = [r*sd[i][0]*np.cos(theta) for theta in np.arange(0,2*np.pi+h,h)]
				y = [r*sd[i][1]*np.sin(theta) for theta in np.arange(0,2*np.pi+h,h)]
				Z = np.transpose(axes[i]) @ np.array([x,y])
				
				plt.plot(Z[0]+x_c,Z[1]+y_c,linewidth=2.5,linestyle='-',c='k')
			
			elif mode=='density':
				# plot elliptical density
				
				for j,r in enumerate(np.linspace(0,sd_max_density,int(sd_max_density/dr))):
					# plot ellipse
					x = [r*sd[i][0]*np.cos(theta) for theta in np.arange(0,2*np.pi+h,h)]
					y = [r*sd[i][1]*np.sin(theta) for theta in np.arange(0,2*np.pi+h,h)]
					Z = np.transpose(axes[i]) @ np.array([x,y])

					plt.gca().fill(Z[0]+x_c,Z[1]+y_c,c='b',alpha=alpha_max*np.exp(-np.sqrt(r)),edgecolor=None)
			elif mode=='arrows':
				# plot cluster center, a single elliptical contour, and arrows along the principal axes
				pass
		
		
	def generate_data(self,add_noise_vars=False, snr=None, p_over_n=None,seed=None,verbose=True):
		"""
		Sample a dataset with clusters. The output data set possesses the
		geometric characteristics specified by this ClusterData object.


		Info on adding noise features
		-----------------------------
		Noise features can be added to the data. The arguments snr and p_over_n
		determine how many noise features are added. 

		The argument snr refers to the desired ratio of meaningful features to
		noise features, whereas the argument p_over_n is the desired ratio of
		the total number of features (p) to the total number of samples (n).
		Thus, snr controls how "sparse" the resulting data is, whereas p_over_n
		determines how high-dimensional the data is (in the statistical sense).

		One or both of snr and p_over_n can be specified. If only one is given,
		add the desired number of noise features exactly. If both snr and
		p_over_n are given, add the least number of noise features that meets or
		exceeds the desired values of snr and p_over_n. Thus, the output data
		is as noisy or noisier than specified by either snr or p_over_n alone.

		
		Parameters
		----------
		self : ClusterData
			The underlying data generator
		add_noise_vars : bool
			If true, add noise features to the data. Need to set snr and/or
			p_over_n to control the amount of noise added. If both snr and
			p_over_n are given, add the least number of noise features that
			meets or exceeds both thresholds (snr and the p_over_n).
		snr : float
			Set the ratio between the number of meaningful features and the
			number of noise features. Only relevant when add_noise_vars=True.
		p_over_n : float
			Set the ratio between the total number of features (p) and the
			number of samples (n). Only relevant when add_noise_vars=True.
		seed : int or None
			Sets the seed for random data generation, if an int is given.

		Returns
		-------
		X : ndarray
			Data matrix whose rows are the data points.
		y : ndarray
			Vector of cluster labels.
		"""

		if verbose: print('Compute class sizes...')
		self.class_sizes = self.class_bal.make_class_sizes(self,seed=seed)

		if verbose: print('Compute covariance structures...')
		axes, sd, cov, cov_inv = self.cov_geom.make_cov(self,seed=seed)
		self.cluster_axes = axes
		self.cluster_sd = sd
		self.cov = cov
		self.cov_inv = cov_inv

		if verbose: print('Place cluster centers...')
		self.centers = (self.center_geom).place_centers(self,seed=seed, verbose=False)

		# compute and report the max and min overlaps beween clusters
		overlap_mat = CenterGeom.overlap_matrix(self.centers,self.cov_inv)
		self.alpha_min_obs = CenterGeom.alpha_min_obs(overlap_mat)
		self.alpha_max_obs = CenterGeom.alpha_max_obs(overlap_mat)

		if verbose: print('Sample data...')
		self.data, self.labels = (self.data_dist).sample(self,seed=seed)

		if verbose: print('Success!')

		if add_noise_vars:
			self.data = self.add_noise(self.data,snr=snr,p_over_n=p_over_n,seed=seed)

		return (self.data, self.labels)


	def add_noise(self, data, snr, p_over_n, model='gauss_mimic', margin=0.1, 
		print_warning=True,seed=None):
		"""
		Add noise features to given data.

		Choosing snr determines how "sparse" the resulting data is, whereas 
		p_over_n determines how high-dimensional the data is. If only one of snr
		and p_over_n is given, add the exact number of noise features required 
		to meet the threshold. If both snr and p_over_n are given, add the least 
		number of noise features that meets or exceeds both thresholds. Thus, the 
		resulting data is as noisy or noisier than required by either threshold.

		Parameters
		----------
		self : ClusterData
			The underlying data generator
		data :
			The data to which noise features are added.
		snr : float
			Set the ratio between the number of meaningful features and the
			number of noise features.
		p_over_n : float
			Set the ratio between the total number of features (p) and the
			number of samples (n).
		model : str
			Determine how noise features are calculated.
		seed : int or None
			Sets the seed for randomized data generation, if an int is given.

		Returns
		-------
		out : ndarray
			Input data with added noise features.

		"""
		np.random.seed(seed)

		n_dim = data.shape[1] 
		n_samples = data.shape[0]

		if (not snr) and (not p_over_n):
			raise Exception('Must provide either snr or p_over_n to determine ' +
							'number of noise features to add.')
		else:
			if snr and not p_over_n:
				# add enough noise features to meet snr threshold
				n_noise_vars = int(np.ceil(n_dim / snr))

			if p_over_n and not snr:
				# add enough noise features to meet p_over_n threshold
				if n_dim/n_samples <= p_over_n:
					n_noise_vars = int(np.ceil(p_over_n * n_samples)) - n_dim
				else:
					noise_vars = []
					if print_warning:
						print('Warning: data already sufficiently high-dimensional,' +
							  ' no noise features added.')
			if snr and p_over_n:
				# add enough noise features to meet both snr and p_over_n thresholds
				n_noise_vars = np.max([int(np.ceil(n_dim/snr)), 
									   int(np.ceil(n_samples*p_over_n))-n_dim])
			
			# create noise
			feature_std = np.std(data,axis=0)
			feature_mean = np.mean(data,axis=0)

			# take a random feature, construct a Gaussian noise feature with the same mean+std
			Z = np.zeros((n,noise_dim))

			for i in range(noise_dim):
				feature_idx = np.random.randint(0,p)
				mean = feature_mean[feature_idx]
				std = feature_std[feature_idx]
				Z[:,i] = np.random.normal(loc=mean,scale=std,size=n)

			# return data with noise features added
			return np.concatenate([data,Z],axis=1)


class CovGeom:
	"""
	Specifies the covariance structure for ClusterData.

	This is largely an abstract superclass. To learn about a concrete
	implementation, see the documentation for maxmin.MaxMinCov.

	Methods
	-------
	__init__(self) : abstract method
	make_cov(self,clusterdata) : abstract method 
	make_orthonormal_axes(self,n_dim,n_axes) : implemented here

	"""
	
	def __init__(self):
		raise NotImplementedError('Cannot instantiate abstract CovGeom.'+
								  ' Choose a provided implementation,'+
								  ' e.g. maxmin.MaxMinCov, or code your'+
								  ' own algorithm for defining cluster'+
								  ' covariance structure.')

	def make_cov(self, clusterdata):
		raise NotImplementedError('Cannot sample covariance structure from'+
								  ' abstract CovGeom. Choose a provided' 
								  ' implementation, e.g. maxmin.MaxMinCov' +
								  ', or code your own algorithm for defining'
								  ' cluster covariance structure.')

	def make_orthonormal_axes(self, n_dim, n_axes, seed=None):
		"""
		Sample orthonormal axes for cluster generation.
		Create n_axes orthonormal axes in n_dim-dimensional space.

		Parameters
		----------
		self : CovGeom
			This instance of CovGeom
		n_dim : int
			Dimensionality of the data space
		n_axes : int
			Number of axes to generate
		seed : int or None
			Sets the seed for randomized axis generation, if an int is given.

		Returns
		-------
		out : (n_axes, n_dim) ndarray
			Axes arranged as a matrix, each row is an axis
		"""
		ortho_matrix = stats.ortho_group.rvs(n_dim,random_state=seed)
		return ortho_matrix[:n_axes, :]



class CenterGeom:
	"""
	Place cluster centers for ClusterData. 

	This is an abstract superclass. To learn about a concrete
	implementation, see the documentation for centers.BoundedSepCenters.

	Methods
	-------
	__init__(self) : abstract method
	make_centers(self, clusterdata) : abstract method
	compute_overlap : 
	overlap_matrix : 

	"""

	def __init__(self):
		raise NotImplementedError('Cannot instantiate abstract CenterGeom.'+
								  ' Choose a provided implementation,'+
								  ' e.g. centers.BoundedSepCenters, or code your'+
								  ' own algorithm for placing cluster centers.')


	def make_centers(self, clusterdata):
		raise NotImplementedError('Cannot sample cluster centers from abstract'+
								  ' CenterGeom. Choose a provided implementation,'+
								  ' e.g. centers.BoundedSepCenters, or code your'+
								  ' own algorithm for placing cluster centers.')


	def compute_overlap(center_1,cov_inv_1,center_2,cov_inv_2):
		"""
		Compute the level alpha of overlap between two clusters.
		"""    
		# compute the mahalanobis distances and calculate the chi2_quantile
		dist1 = mahalanobis(center_1,center_2,cov_inv_1)
		dist2 = mahalanobis(center_1,center_2,cov_inv_2)
		chi2_quantile = (1/((1/dist1) + (1/dist2)))**2

		# compute probability P(chi2 <= chi2_quantile) and get alpha
		df = len(center_1)
		alpha = 1 - (stats.chi2(df=df).cdf(chi2_quantile))

		return alpha
	

	def overlap_matrix(centers,cov_inv):
		"""
		Compute the matrix of overlaps between pairs of clusters.
		"""
		n_centers = centers.shape[0]
		pw_overlap = np.zeros((n_centers,n_centers))

		for i in range(n_centers):
			for j in range(i+1,n_centers):
				pw_overlap[i,j] = CenterGeom.compute_overlap(centers[i,:], cov_inv[i],
																	 centers[j,:], cov_inv[j])
				#print('overlap', pw_overlap[i,j], ' by ', i,j)

		pw_overlap = pw_overlap + np.transpose(pw_overlap)

		return pw_overlap

	def alpha_max_obs(overlap_mat):
		"""
		Compute maximum overlap between pairs of cluster centers.
		"""
		return np.max(overlap_mat.flatten())

	def alpha_min_obs(overlap_mat):
		"""
		Compute the minimax overlap between cluster centers. That is,
		for each cluster center, compute the maximum overlap (overlap with
		the nearest other center) and take the minimum of these overlaps.
		The resulting metric quantifies the degree of isolation of the most
		isolated cluster.
		"""
		n_centers = overlap_mat.shape[0]
		Z = overlap_mat.copy()
		#Z[np.arange(n_centers),np.arange(n_centers)] = np.Inf

		return np.min(np.max(Z,axis=1))


class ClassBal:
	"""
	Sample class sizes (number of data points in each cluster) for ClusterData.

	This is an abstract superclass. To learn about a concrete
	implementation, see the documentation for maxmin.MaxMinBal.

	Methods
	-------
	__init__(self) : abstract method
	make_class_sizes(self, clusterdata) : abstract method

	"""

	def __init__(self):
		raise NotImplementedError('Cannot instantiate abstract ClassBal.' + 
								  ' Choose a provided implementation, e.g.'+
								  'maxmin.MaxMinBal, or code your own '+
								  'algorithm for sampling class sizes.')

	def make_class_sizes(self, clusterdata):
		raise NotImplementedError('Cannot sample class sizes from abstract'+
								  ' ClassBal. Choose a provided implementation,'+
								  ' e.g. maxmin.MaxMinBal, or code your own '+
								  'algorithm for sampling class sizes.')


class DataDist:
	"""
	Specify data distribution for ClusterData.

	This is largely an abstract superclass. To learn about concrete
	implementations, see the documentations for distributions.GaussianDist,
	distributions.ExpDist, or distributions.tDist.

	Methods
	-------
	__init__(self) : abstract method
	sample(self, clusterdata) : implemented here
	sample_cluster(self, cluster_size, mean, axes, sd) : abstract method

	"""

	def __init__(self):
		raise NotImplementedError('Cannot instantiate abstract DataDist. Choose '+
								  ' a provided implementation, e.g. GaussianDist,' +
								  ' or code your own data distribution.')

	def sample(self, clusterdata, seed=None):
		n_clusters = clusterdata.n_clusters
		n_samples = clusterdata.n_samples
		n_dim = clusterdata.n_dim
		class_sizes = clusterdata.class_sizes
		centers = clusterdata.centers

		axes = clusterdata.cluster_axes
		sd = clusterdata.cluster_sd

		X = np.full(shape=(n_samples, n_dim), fill_value=np.nan)
		y = np.full(n_samples, fill_value=np.nan).astype(int)

		start = 0
		for i in range(n_clusters):
			end = start + class_sizes[i]
			# Set class label
			y[start:end] = i
			# Sample data
			X[start:end,:] = self.sample_cluster(class_size=class_sizes[i], 
										  mean=centers[i], axes=axes[i],
										  sd=sd[i],seed=seed+i if seed else seed)
			start = end

		return (X, y)

	def sample_cluster(self, cluster_size, mean, axes, sd, seed=None):
		raise NotImplementedError('Cannot sample cluster from abstract DataDist. '+
								  'Choose a provided implementation, e.g. '
								  'GaussianDist, or code your own data distribution.')


import numpy as np
import scipy.stats as stats
from .maxmin import MaxMinClusters
from .centers import FastHighDimCenters

class TestStream():
    """
    Iterator that randomly generates synthetic data.
    """
    
    def __init__(self,n_tests=20,seed=None,return_labels=False, sample_dim=lambda: np.random.exponential(25)):
        """
        Instantiate TestStream object.
        """
        self.n_tests = n_tests
        self.iter_count = 0
        self.return_labels = return_labels
        self.seed = seed
        self.sample_dim = sample_dim
        
    def __iter__(self):
        """
        Return iterator.
        """
        return self
        
    def __next__(self):
        """
        Sample next data set.
        """
        if (self.iter_count == self.n_tests):
            raise StopIteration
        
        clusterdata = MaxMinClusters()
        self.randomize_parameters(clusterdata, 
                                        seed=(None if not self.seed else self.seed + 4783*self.iter_count) % 7727)
        
        if (clusterdata.n_dim >= 20):
            clusterdata.center_geom = FastHighDimCenters()

        X, y = clusterdata.generate_data(verbose=False, 
                                         seed=(None if not self.seed else self.seed + 3529*self.iter_count) % 7727)
        self.iter_count += 1
        
        attr_dict = {'k': clusterdata.n_clusters, 'p': clusterdata.n_dim, 'n': clusterdata.n_samples,
                     'alpha_max': clusterdata.alpha_max_obs, 'alpha_min': clusterdata.alpha_min_obs}
        
        if self.return_labels:
            return (X,attr_dict,y)
        else:
            return (X,attr_dict)
    
    def randomize_parameters(self, clusterdata, seed=None, max_k=30, min_pts_per_cluster=50, 
                             max_sample_sz=2000):
        """
        Randomize parameters of ClusterData object.
        """
        if seed:
            np.random.seed(seed)
            
        clusterdata.n_clusters = np.random.choice(a=np.arange(2,max_k+1))
        clusterdata.n_samples = np.max([clusterdata.n_clusters*min_pts_per_cluster, 
                                        100*np.random.choice(np.arange(1,int(max_sample_sz/100)))])
        clusterdata.n_dim = np.max([2, int(self.sample_dim())]) #int(np.random.exponential(ref_dim))
        clusterdata.alpha_max = np.random.uniform(0.01,0.2)
        clusterdata.alpha_min = 1e-5
        clusterdata.aspect_ref = np.random.uniform(1,5)
        clusterdata.radius_maxmin = np.random.uniform(1,5)
        clusterdata.aspect_maxmin = np.random.uniform(1,5)
        clusterdata.imbal_maxmin = np.random.uniform(1,5)


"""
This module provides algorithms for sampling cluster centers for a 
ClusterData object.

CLASSES AND METHODS

	BoundedSepCenters : sample cluster centers obeying separation constraints
		__init__(self, min_sep, max_sep, ...)
		place_centers(self, clusterdata)
		
"""

import numpy as np
import scipy.stats as stats
from .core import CenterGeom
from scipy.spatial.distance import mahalanobis, pdist, squareform
from scipy.stats import chi2
from numpy.linalg import qr




class BoundedSepCenters(CenterGeom):
	"""
	Sample cluster centers by specifying maximum and minimum separation 
	between centers.

	The minimum separation constraint ensures that no clusters are too 
	close together. On the other hand, the maximum separation constraint 
	guarantees that no cluster is too far away from all other clusters

	Separation is expressed on a scale derived from Mahalanobis distance. 
	A separation of 1.0 means that two clusters are just touching (when 
	each cluster is considered as a hard sphere filling 95% of the 
	cluster's probability mass). Lower values of separation imply overlap 
	between the clusters; higher values cleanly separate the clusters. 

	Note that the above interpretation of separation value is strictly 
	valid only for Gaussian clusters; for clusters with other data 
	distributions, separation values below or above 1.0 may be necessary
	for two clusters to be "just touching."

	Attributes
	----------
	min_sep : float (positive)
		Minimum separation between clusters. Value 1.0 means "just touching",
		lower values allow overlaps. See discussion in the class description.
	max_sep : float (positive)
		Maximum separation between clusters. Value 1.0 means "just touching",
		lower values imply overlap. See discussion in the class description.
	packing : float between 0 and 1, default=0.1
		Ratio of total cluster volume to sampling volume, used in rejection
		sampling. Here, cluster volume is computed from the maximum standard
		deviation among cluster shapes.

	"""
	
	def __init__(self, alpha_min, alpha_max, packing=.1):
		"""
		Create a BoundedSepCenters object.

		Parameters
		----------
		min_sep : float (positive)
			Minimum separation between clusters. Value 1.0 means "just touching",
			lower values allow overlaps. See discussion in the class description.
			max_sep : float (positive)
		Maximum separation between clusters. Value 1.0 means "just touching",
			lower values imply overlap. See discussion in the class description.
		packing : float between 0 and 1, default=0.1
			Ratio of total cluster volume to sampling volume, used in rejection
			sampling. Here, cluster volume is computed from the maximum standard
			deviation among cluster shapes.

		Returns
		-------
		out : BoundedSepCenters
			Object for sampling cluster centers obeying separation constraints.
		"""
		self.alpha_min = alpha_min
		self.alpha_max = alpha_max
		self.packing = packing
		

	def place_centers(self, clusterdata, seed=None, verbose=True):
		"""
		Place cluster centers sequentially with rejection sampling.

		Sample possible cluster centers uniformly within a box. To achieve 
		desired separation between clusters, only accept new centers whose
		separation from all other clusters is greater than self.min_sep (ensure
		no clusters overlap too much). In addition, only accept new centers
		whose separation from at least one other cluster is less than 
		self.max_sep (ensure no cluster is far away from all other clusters).

		Parameters
		----------
		self : BoundedSepCenters
			This instance of BoundedSepCenters
		clusterdata : ClusterData
			The data generator

		Returns
		-------
		centers : ndarray
			Cluster centers arranged in a matrix. Each row is a cluster center.
		"""

		np.random.seed(seed)

		n_clusters = clusterdata.n_clusters
		n_dim = clusterdata.n_dim

		# find the maximum sd
		cluster_sd = clusterdata.cluster_sd
		max_sd = 2*np.prod(np.array([(np.prod(sd_vec))**(1/n_dim) for sd_vec in cluster_sd]))**(1/n_dim)
		#max_sd = np.max(np.array([np.max(sd_vec) for sd_vec in cluster_sd]))

		cov_inv = clusterdata.cov_inv
		
		centers = np.zeros(shape=(n_clusters, n_dim))

		total_cluster_vol = n_clusters * ((4*max_sd)**n_dim)
		sampling_vol = total_cluster_vol / self.packing
		sampling_width = sampling_vol**(1/n_dim)

		# compute reference chi square value (take 0.95 quantile)
		q_chi_sq_repel = chi2.ppf(1-self.alpha_max, df=n_dim)
		q_chi_sq_attract = chi2.ppf(1-self.alpha_min, df=n_dim)
		
		for new_center in range(n_clusters):
			accept = False
			while not accept:
				proposed_center = sampling_width * np.random.uniform(size=n_dim)
				far_enough = True # Need to uphold minimum distance to other centers.
				close_enough = False # New center shouldn't be far away from all centers.

				# Check distances to previously selected centers
				for prev_ctr in range(new_center):
					d_1 = mahalanobis(proposed_center, centers[prev_ctr], 
										 cov_inv[new_center])
					d_2 = mahalanobis(proposed_center, centers[prev_ctr], 
										 cov_inv[prev_ctr])

					# quantile corresponding to chi^2 distribution with df=n_dim
					q_sq = 1/(((1/d_1) + (1/d_2))**2)
					# min_sep = q_sq_(1-alpha_min), max_sep = q_sq_(1-alpha_max),
					# e.g. min_sep = 0.95, max_sep = 0.9999

					if (q_sq < q_chi_sq_repel):
						far_enough = False 
						break

					if (q_sq < q_chi_sq_attract):
						close_enough = True
				
				if ((far_enough and close_enough) or (new_center == 0)):
					accept = True
					#UNCOMMENT FOR DEBUGGING PURPOSES:
					#if not (new_center == 0): 
					#	print(chi_1, chi_2)
					centers[new_center,:] = proposed_center
					if verbose: print('\t' + str(1+new_center) + 
						'/' + str(n_clusters) + ' placed!')
				
		return(centers)



class FastHighDimCenters(CenterGeom):
	"""
	Sample cluster centers fast for data sets with high intrinsic dimensionality.
	"""
	
	def __init__(self, alpha_min=1e-5,alpha_max=0.05, 
				 packing_min=1e-2, packing_max=2,thold=1.05, max_runs=5):
		"""
		Instantiate a FastHighDimCenters object.
		"""
		self.max_runs = max_runs
		self.log_thold = np.log(thold)
		self.log_packing_min = np.log(packing_min)
		self.log_packing_max = np.log(packing_max)
		self.alpha_min = alpha_min
		self.alpha_max = alpha_max
		self.alpha_min_obs = None
		self.alpha_max_obs = None
	
	def place_centers(self, clusterdata, seed=None, verbose=False):
		"""
		Sample cluster centers repeatedly for different values of the
		packing parameter. Take the centers that best fit the desired
		minimum and maximum overlap values.
		"""
		# make adjustments for the dimensionality
		self.log_packing_min = (self.log_packing_min)*(clusterdata.n_dim/2)
		self.log_packing_max = (self.log_packing_max)*(clusterdata.n_dim/2)
		self.log_thold = (self.log_thold)*clusterdata.n_dim*2
		
		self.binary_search_packing(clusterdata,self.log_packing_min,self.log_packing_max,
								   attempts_remaining=self.max_runs,
								   verbose=verbose)
		return clusterdata.centers
		
	def overlap_error(alpha_min_obs,alpha_max_obs,alpha_min_desired,alpha_max_desired,
					  inf_proxy=100,verbose=True):
		"""
		Compute how far observed overlaps are outside the desired box [alpha_min,alpha_max]
		in log space and penalize the overhang linearly.
		"""
		if verbose:
			print('alpha_min_obs',alpha_min_obs,'alpha_max_obs',alpha_max_obs)

		if (alpha_min_obs == 0):
			diff2 = inf_proxy
		else:
			diff2 = np.log(alpha_min_desired) - np.log(alpha_min_obs)

		if (alpha_max_obs == 0):
			diff1 = inf_proxy
		else:
			diff1 = np.log(alpha_max_obs) - np.log(alpha_max_desired)
		
		return diff1*(diff1>0) + diff2*(diff2>0)
		
	def binary_search_packing(self, clusterdata, log_packing_min, 
							  log_packing_max, attempts_remaining,
							  n_restarts=5,verbose=True):
		"""
		"""
		avg_log_packing = (log_packing_min+log_packing_max)/2
		
		if verbose:
			print('log packing ->', avg_log_packing)
		
		seed_list = np.random.choice(n_restarts*1000,size=n_restarts,replace=False)
		# randomly choose cluster centers with current packing parameter
		min_error = np.Inf
		min_idx = None

		cov_save_list = []
		
		for i in range(n_restarts):
			# place cluster centers
			cov_save_list.append((clusterdata.cov,clusterdata.cov_inv, 
									clusterdata.cluster_axes, clusterdata.cluster_sd))
			self.attempt_to_place_centers(avg_log_packing, clusterdata, seed=seed_list[i],
										  modify_clusterdata=True)
			
			# report error
			overlap_mat = CenterGeom.overlap_matrix(clusterdata.centers,
															clusterdata.cov_inv)
			alpha_min = CenterGeom.alpha_min_obs(overlap_mat)
			alpha_max = CenterGeom.alpha_max_obs(overlap_mat)
			error = FastHighDimCenters.overlap_error(alpha_min_obs=alpha_min,
													 alpha_max_obs=alpha_max,
													 alpha_min_desired=self.alpha_min,
													 alpha_max_desired=self.alpha_max,
													 verbose=verbose)
			# update best error
			if (error < min_error):
				min_idx = i
				min_error = error
			
		# continue binary search
		if ((log_packing_max - log_packing_min) >= self.log_thold) and (attempts_remaining>0):
			cov_save_list = []
			if (alpha_max > self.alpha_max):
				# go left -> reduce packing
				self.binary_search_packing(clusterdata,log_packing_min,avg_log_packing,
										   n_restarts=n_restarts,
										   attempts_remaining=attempts_remaining-1,verbose=verbose)
			else:
				# go right -> increase packing
				self.binary_search_packing(clusterdata,avg_log_packing,log_packing_max,
										   n_restarts=n_restarts,
										   attempts_remaining=attempts_remaining-1,verbose=verbose)
		else:
			clusterdata.modify_cov_structure(*cov_save_list[min_idx])
			self.attempt_to_place_centers(log_packing=avg_log_packing, clusterdata=clusterdata, 
												 seed=seed_list[min_idx], verbose=verbose,
												 modify_clusterdata=True)
			# recompute overlap
			overlap_mat = CenterGeom.overlap_matrix(clusterdata.centers,
															clusterdata.cov_inv)
			self.alpha_min_obs = CenterGeom.alpha_min_obs(overlap_mat)
			self.alpha_max_obs = CenterGeom.alpha_max_obs(overlap_mat)
			
			# report packing
			self.log_packing = avg_log_packing
		
	
	def attempt_to_place_centers(self, log_packing, clusterdata, seed=None, verbose=True,
								 modify_clusterdata=True):
		"""
		Sample cluster centers and adjust covariance matrices.
		"""
		if seed:
			np.random.seed(seed)
		
		n_clusters = clusterdata.n_clusters
		n_dim = clusterdata.n_dim

		# find reference sd for the whole data
		cluster_sd = clusterdata.cluster_sd
		log_ref_sd = np.mean(np.array([np.mean(np.log(sd_vec)) for sd_vec in cluster_sd]))

		centers = np.zeros(shape=(n_clusters, n_dim))
	
		#total_cluster_vol = n_clusters * ((4*ref_sd)**n_dim)
		#print('total cluster vol', total_cluster_vol)
		#sampling_vol = total_cluster_vol / self.packing
		#sampling_width = sampling_vol**(1/n_dim)
		
		log_total_cluster_vol = n_dim*(np.log(4) + log_ref_sd)
		log_sampling_vol = log_total_cluster_vol - log_packing
		sampling_width = np.exp(log_sampling_vol/n_dim)

		centers = sampling_width * np.random.uniform(size=(n_clusters,n_dim))
		
		# adjust covariance structures to accommodate the new centers
		if modify_clusterdata:
			clusterdata.centers = centers
			FastHighDimCenters.adjust_cov(clusterdata)
		
		return centers

	
	def adjust_cov(clusterdata):
		"""
		Adjust covariance matrices after sampling cluster centers.
		"""

		# compute pairwise distances between cluster centers
		center_array = clusterdata.centers
		center_pw_dist = squareform(pdist(center_array))
		center_pw_dist[np.arange(center_pw_dist.shape[0]),np.arange(center_pw_dist.shape[0])] = np.Inf
		

		# for each cluster center, compute minimum distances to the nearest other center
		min_dist = np.min(center_pw_dist,axis=1)

		# compute volume for each covariance matrix
		vol_array = np.zeros(clusterdata.n_clusters)
		for k in range(clusterdata.n_clusters):
			vol_array[k] = np.prod(4*clusterdata.cluster_sd[k])

		# sort covariance matrices by volume, increasing order
		vol_order = np.argsort(vol_array)
		# sort cov in order of minimum distance to nearest other center, increasing order
		cov_order = np.argsort(np.argsort(min_dist)[np.arange(clusterdata.n_clusters)])

		new_cov_list = []
		new_cov_inv_list = []
		new_sd_list = []
		new_axes_list = []
		# assign smallest-volume covariance matrix to smallest-min-distance centers
		for k in range(clusterdata.n_clusters):
			#print('\n\nDoing center ', str(k), '...')
			ctr = center_array[k,:]
			cov = clusterdata.cov[vol_order[cov_order[k]]]
			axes = clusterdata.cluster_axes[vol_order[cov_order[k]]] 
			axes_sd = clusterdata.cluster_sd[vol_order[cov_order[k]]]

			# find vector corresponding to minimum distance, and normalize
			closest_ctr_idx = np.argmin(center_pw_dist[k,:])
			#print('closest other center:', closest_ctr_idx)
			min_dist_vec = center_array[closest_ctr_idx,:] - ctr
			min_dist_vec = min_dist_vec/np.sqrt(np.dot(min_dist_vec,min_dist_vec))
			#print('minimum distance vector:', min_dist_vec)

			# find longest principal axis
			longest_axis_idx = np.argmax(axes_sd)
			longest_axis = axes[longest_axis_idx,:]
			#print('longest axis:', longest_axis)

			# project longest axis away from the min dist vector
			projection_coef = np.dot(min_dist_vec,longest_axis)
			#print('proj coef:', projection_coef)
			new_longest_axis = longest_axis - projection_coef*min_dist_vec 
			#print('new longest axis:', new_longest_axis)

			# Step 1: replace longest axis by the new longest axis, place longest axis first
			# Comment: placing longest axis first is important for QR decomposition-- otherwise,
			# the decomposition will not leave the longest axis intact.
			axes[longest_axis_idx,:] = axes[0,:]
			axes[0,:] = new_longest_axis

			# Step 2: calculate QR decomp of the resulting matrix, use as new axes
			new_axes = np.transpose(qr(np.transpose(axes))[0])

			# Step 3: put together new cov matrix from new axes and old standard deviations
			axes_sd = np.sort(axes_sd)[::-1]
			cov = np.transpose(new_axes) @ np.diag(axes_sd**2) @ new_axes
			
			cov_inv = np.transpose(new_axes) @ np.diag(axes_sd**(-2)) @ new_axes

			# Step 4: reset cov matrix in cluster data
			new_cov_list.append(cov)
			new_cov_inv_list.append(cov_inv)
			new_sd_list.append(axes_sd)
			new_axes_list.append(new_axes)
			
		clusterdata.modify_cov_structure(new_cov_list, new_cov_inv_list, new_axes_list, new_sd_list)